"""Domain client for registry Domain service."""
from datetime import datetime
from enum import Enum, IntEnum, unique
from typing import Any, AsyncIterable, Dict, Mapping, Optional, Sequence, Set, Type, Union, cast

import fred_api.registry.domain.service_admin_grpc_pb2_grpc
import fred_api.registry.domain.service_domain_grpc_pb2_grpc
import grpc
from aioitertools.more_itertools import chunked
from aioitertools.types import AnyIterable
from fred_api.registry.domain.domain_history_types_pb2 import DomainHistoryReply, DomainHistoryRequest
from fred_api.registry.domain.domain_info_types_pb2 import (
    DomainIdReply,
    DomainIdRequest,
    DomainInfoReply,
    DomainInfoRequest,
)
from fred_api.registry.domain.domain_life_cycle_stage_types_pb2 import DomainLifeCycleStageRequest
from fred_api.registry.domain.domain_state_history_types_pb2 import DomainStateHistoryReply, DomainStateHistoryRequest
from fred_api.registry.domain.domain_state_types_pb2 import DomainStateReply, DomainStateRequest
from fred_api.registry.domain.domains_by_contact_types_pb2 import DomainsByContactReply, DomainsByContactRequest
from fred_api.registry.domain.service_admin_grpc_pb2 import (
    BatchDeleteDomainsReply,
    BatchDeleteDomainsRequest,
    DomainContactInfo,
    GetDomainsDeleteNotifyInfoRequest,
    GetDomainsOutzoneNotifyInfoRequest,
    ManageDomainStateFlagsReply,
    ManageDomainStateFlagsRequest,
    UpdateDomainsAdditionalNotifyInfoReply,
    UpdateDomainsDeleteAdditionalNotifyInfoRequest,
    UpdateDomainsOutzoneAdditionalNotifyInfoRequest,
)
from fred_types import BaseModel, ContactId, ContactRef, DomainHistoryId, DomainId, DomainRef, KeysetRef, NssetRef
from frgal.aio import AsyncGrpcClient
from google.protobuf.message import Message

from .common import DatetimeRange, ObjectEvents, RegistryDecoder, set_request_pagination
from .exceptions import ContactDoesNotExist, DomainDoesNotExist, InvalidHistoryInterval
from .object import RegistryObjectClient

UpdateDomainsNotifyInfoRequest = Union[
    UpdateDomainsDeleteAdditionalNotifyInfoRequest,
    UpdateDomainsOutzoneAdditionalNotifyInfoRequest,
]


class Domain(BaseModel):
    """Domain info.

    Attributes:
        expiration_warning_scheduled_at: An expected date of an expiration warning for the domain.
        outzone_unguarded_warning_scheduled_at: An expected date of an outzone warning for the domain.
        outzone_scheduled_at: An expected date of the domain removal from the zone.
        delete_warning_scheduled_at: An expected date of a deletion warning for the domain.
        delete_candidate_scheduled_at: An expected date of the domain deletion.
        validation_expires_at: Timestamp of validation expiration for ENUM domains. Expected date if in future.
        outzone_at: Actual date of the domain removal from the zone. None if domain is in the zone.
        delete_candidate_at: Actual date of the domain deletion. None if domain is not deleted.
    """

    domain_id: DomainId
    fqdn: str
    domain_history_id: Optional[DomainHistoryId] = None
    registrant: Optional[ContactRef] = None
    sponsoring_registrar: Optional[str] = None
    nsset: Optional[NssetRef] = None
    keyset: Optional[KeysetRef] = None
    administrative_contacts: Sequence[ContactRef] = []
    auth_info: Optional[str] = None
    events: Optional[ObjectEvents] = None
    # Domain life cycle attributes.
    expires_at: Optional[datetime] = None
    outzone_at: Optional[datetime] = None
    delete_candidate_at: Optional[datetime] = None
    expiration_warning_scheduled_at: Optional[datetime] = None
    outzone_unguarded_warning_scheduled_at: Optional[datetime] = None
    outzone_scheduled_at: Optional[datetime] = None
    delete_warning_scheduled_at: Optional[datetime] = None
    delete_candidate_scheduled_at: Optional[datetime] = None
    validation_expires_at: Optional[datetime] = None


@unique
class DomainContactRole(str, Enum):
    """Role of contact linked to domain."""

    HOLDER = "holder"
    ADMIN = "admin"
    KEYSET_TECH = "keyset_tech_contact"
    NSSET_TECH = "nsset_tech_contact"


@unique
class NotDeletedReason(IntEnum):
    """Enum with reasons for domain not to be deleted."""

    UNDEFINED = 0
    DOMAIN_BLOCKED = 1
    DOMAIN_DOES_NOT_EXIST = 2
    DOMAIN_HISTORY_ID_MISMATCH = 3
    INSUFFICIENT_PERMISSIONS = 4


@unique
class ContactInfoType(str, Enum):
    """Contact information type."""

    EMAIL = "email"
    TELEPHONE = "telephone"


class ContactInfo(BaseModel):
    """Contact information."""

    type: ContactInfoType
    value: str


class DomainDecoder(RegistryDecoder):
    """Domain service decoder."""

    def __init__(self) -> None:
        super().__init__()
        self.set_decoder(DomainsByContactReply.Data.Domain, self._decode_domain_by_contact)
        self.set_exception_decoder(DomainIdReply.Exception.DomainDoesNotExist, DomainDoesNotExist)
        self.set_exception_decoder(DomainInfoReply.Exception.DomainDoesNotExist, DomainDoesNotExist)
        self.set_exception_decoder(DomainHistoryReply.Exception.DomainDoesNotExist, DomainDoesNotExist)
        self.set_exception_decoder(DomainHistoryReply.Exception.InvalidHistoryInterval, InvalidHistoryInterval)
        self.set_exception_decoder(DomainStateReply.Exception.DomainDoesNotExist, DomainDoesNotExist)
        self.set_exception_decoder(DomainStateHistoryReply.Exception.DomainDoesNotExist, DomainDoesNotExist)
        self.set_exception_decoder(DomainStateHistoryReply.Exception.InvalidHistoryInterval, InvalidHistoryInterval)
        self.set_exception_decoder(DomainsByContactReply.Exception.ContactDoesNotExist, ContactDoesNotExist)
        self.set_exception_decoder(DomainsByContactReply.Exception.InvalidOrderBy, ValueError)

    def _decode_domain_by_contact(self, value: DomainsByContactReply.Data.Domain) -> Dict[str, Any]:
        """Transform booleans to DomainContactRole set."""
        roles = set()
        if value.is_holder:
            roles.add(DomainContactRole.HOLDER)
        if value.is_admin_contact:
            roles.add(DomainContactRole.ADMIN)
        if value.is_keyset_tech_contact:
            roles.add(DomainContactRole.KEYSET_TECH)
        if value.is_nsset_tech_contact:
            roles.add(DomainContactRole.NSSET_TECH)
        return {
            "domain": DomainRef(id=self.decode(value.domain.id), fqdn=self.decode(value.domain.fqdn)),
            "domain_history_id": DomainHistoryId(self.decode(value.domain_history_id)),
            "roles": roles,
            "is_deleted": value.is_deleted,
        }


class DomainClient(RegistryObjectClient[DomainId, DomainHistoryId]):
    """Domain service gRPC client."""

    service = "Domain"
    object_id = "domain_id"
    object_history_id = "domain_history_id"
    object_handle = "fqdn"

    def __init__(self, netloc: str, credentials: Optional[grpc.ChannelCredentials] = None) -> None:
        """Initialize domain client instance.

        Args:
            netloc: Netloc of Domain service server.
            credentials: Optional credentials.
        """
        super().__init__(
            netloc,
            decoder=DomainDecoder(),
            grpc_modules=[fred_api.registry.domain.service_domain_grpc_pb2_grpc],
            credentials=credentials,
        )

    async def get_domain_id(self, fqdn: str) -> DomainId:
        """Get domain id.

        Args:
            fqdn: Domain FQDN.

        Raises:
            DomainDoesNotExist: If domain does not exist.
        """
        return DomainId(await self.get_object_id(request=DomainIdRequest(), method="get_domain_id", handle=fqdn))

    async def get_domain_info(self, domain_id: DomainId, history_id: Optional[DomainHistoryId] = None) -> Domain:
        """Get domain info.

        Args:
            domain_id: Domain ID.
            history_id: Domain history ID.

        Raises:
            DomainDoesNotExist: If domain does not exist.
        """
        result = await self.get_object_info(
            request=DomainInfoRequest(),
            method="get_domain_info",
            object_id=domain_id,
            history_id=history_id,
        )
        result.pop("registrant_old", None)
        result.pop("nsset_old", None)
        result.pop("keyset_old", None)
        result.pop("administrative_contacts_old", None)
        return Domain(**result)

    async def get_domain_state(self, domain_id: DomainId) -> Dict[str, bool]:
        """Get domain state consisting of state flags and their boolean values.

        Raises:
            DomainDoesNotExist: If domain does not exist.

        Args:
            domain_id: Domain ID.
        """
        return await self.get_object_state(request=DomainStateRequest(), method="get_domain_state", object_id=domain_id)

    async def get_domain_history(
        self,
        domain_id: DomainId,
        start: Optional[Union[DomainHistoryId, datetime]] = None,
        end: Optional[Union[DomainHistoryId, datetime]] = None,
    ) -> Dict[str, Any]:
        """Get domain history timeline.

        Args:
            domain_id: Domain ID.
            start: Datetime or history_id of history interval start.
            end: Datetime or history_id of history interval end.

        Raises:
            DomainDoesNotExist: If domain does not exist.
            InvalidHistoryInterval: In case of invalid history interval.
        """
        return await self.get_object_history(
            request=DomainHistoryRequest(),
            method="get_domain_history",
            object_id=domain_id,
            start=start,
            end=end,
        )

    async def get_domain_state_history(
        self,
        domain_id: DomainId,
        start: Optional[Union[DomainHistoryId, datetime]] = None,
        end: Optional[Union[DomainHistoryId, datetime]] = None,
    ) -> Dict[str, Any]:
        """Get domain state history timeline.

        Args:
            domain_id: Domain ID.
            start: Datetime or history_id of history interval start.
            end: Datetime or history_id of history interval end.

        Raises:
            DomainDoesNotExist: If domain does not exist.
            InvalidHistoryInterval: In case of invalid history interval.
        """
        return await self.get_object_state_history(
            request=DomainStateHistoryRequest(),
            method="get_domain_state_history",
            object_id=domain_id,
            start=start,
            end=end,
        )

    async def get_domains_by_contact(
        self,
        contact_id: ContactId,
        roles: Set[DomainContactRole],
        *,
        include_deleted: bool = False,
        order_by: Optional[str] = None,
        fqdn_filter: Optional[str] = None,
        page_token: Optional[str] = None,
        page_size: Optional[int] = None,
    ) -> Dict[str, Any]:
        """Get domains linked to contact.

        Args:
            contact_id: Contact ID.
            roles: Set of roles in which is contact linked to domain.
            include_deleted: Include deleted domains that were linked to contact in moment of their deletion.
            order_by: Name of column to order results by.
                Can be prefixed with `+` (asc) or `-` (desc). `+` is optional.
            fqdn_filter: Filter for domain names.
                `*` means any sequence of characters.
                Pattern is matched against the whole domain name.
            page_token: Page token for pagination.
            page_size: Page size for pagination.

        Returns:
            List of domain ids and their appropriate contact roles.

        Raises:
            ContactDoesNotExist: If contact does not exist.
            ValueError: If order_by column name is invalid.
        """
        request = DomainsByContactRequest()
        request.contact_id.uuid.value = contact_id
        set_request_pagination(request, page_token=page_token, page_size=page_size)

        if DomainContactRole.HOLDER in roles:
            request.include_holder = True
        if DomainContactRole.ADMIN in roles:
            request.include_admin_contact = True
        if DomainContactRole.KEYSET_TECH in roles:
            request.include_keyset_tech_contact = True
        if DomainContactRole.NSSET_TECH in roles:
            request.include_nsset_tech_contact = True

        request.include_deleted_domains = include_deleted
        if order_by:
            request.order_by = order_by
        if fqdn_filter:
            request.fqdn_filter = fqdn_filter

        result: Dict[str, Any] = {"domains": []}
        async for message in self.call_stream(self.service, "get_domains_by_contact", request):  # pragma: no cover
            if pagination := message.get("pagination"):
                result.setdefault("pagination", pagination)
            result["domains"] = message.get("domains", [])

        return result

    async def get_domain_life_cycle_stage(self, fqdn: str) -> Optional[DomainId]:
        """Return result based on domain status.

        Args:
            fqdn: Domain FQDN.

        Returns:
            * Domain ID in case the domain is registered.
            * None if domain is in delete candidate state or have been deleted as such today.

        Raises:
            DomainDoesNotExist: If domain is not registered.
        """
        request = DomainLifeCycleStageRequest()
        request.fqdn.value = fqdn
        result = await self.call(self.service, "get_domain_life_cycle_stage", request=request)
        if "domain_id" in result:
            return cast(DomainId, result["domain_id"])
        elif "delete_candidate" in result:
            return None
        else:
            raise DomainDoesNotExist


class DomainAdminDecoder(RegistryDecoder):
    """Domain Admin service decoder."""

    def __init__(self) -> None:
        super().__init__()
        self.set_decoder(DomainContactInfo, self._decode_domain_contact_info)
        self.set_decoder(BatchDeleteDomainsReply.NotDeleted, self._decode_not_deleted)

        self.set_exception_decoder(ManageDomainStateFlagsReply.Exception.DomainDoesNotExist, DomainDoesNotExist)
        self.set_exception_decoder(ManageDomainStateFlagsReply.Exception.DomainStateFlagRequestDoesNotExist, ValueError)
        self.set_exception_decoder(ManageDomainStateFlagsReply.Exception.DomainStateFlagDoesNotExist, ValueError)
        self.set_exception_decoder(
            UpdateDomainsAdditionalNotifyInfoReply.Exception.DomainDoesNotExist, DomainDoesNotExist
        )
        self.set_exception_decoder(
            UpdateDomainsAdditionalNotifyInfoReply.Exception.InvalidDomainContactInfo, ValueError
        )

    def _decode_domain_contact_info(self, value: DomainContactInfo) -> Optional[ContactInfo]:
        if value.HasField("email"):
            return ContactInfo(type=ContactInfoType.EMAIL, value=value.email)
        elif value.HasField("telephone"):
            return ContactInfo(type=ContactInfoType.TELEPHONE, value=value.telephone)
        else:
            return None

    def _decode_not_deleted(self, value: BatchDeleteDomainsReply.NotDeleted) -> dict:
        decoded = cast(Dict[str, Any], self._decode_message(value))
        decoded["reason"] = NotDeletedReason(decoded["reason"])
        return decoded


class DomainAdminClient(AsyncGrpcClient):
    """Domain Admin service gRPC client."""

    service = "Admin"
    decoder_cls = DomainAdminDecoder
    grpc_modules = (fred_api.registry.domain.service_admin_grpc_pb2_grpc,)

    @staticmethod
    async def _make_batch_delete_domains_requests(
        domains: AnyIterable[Domain], *, batch_size: int = 100
    ) -> AsyncIterable[BatchDeleteDomainsRequest]:
        """Request generator for `batch_delete_domains` method."""
        async for chunk in chunked(domains, batch_size):
            request = BatchDeleteDomainsRequest()
            for domain in chunk:
                message = BatchDeleteDomainsRequest.Domain()
                message.domain_id.uuid.value = domain.domain_id
                if domain.domain_history_id is None:
                    raise ValueError("{!r} is missing a domain_history_id")
                message.domain_history_id.uuid.value = domain.domain_history_id
                request.domains.append(message)
            yield request

    async def batch_delete_domains(self, domains: AnyIterable[Domain], *, batch_size: int = 100) -> AsyncIterable[Dict]:
        """Batch deletion of domains.

        Args:
            domains: Iterable (or async iterable) of domains to be deleted.
            batch_size: Number of domains in single stream request.

        Returns:
            Async iterable of dictionaries with following keys:
                domain_id: Domain identification (as a string).
                reason: Reason for not deleting domain (as a NotDeletedReason enumeration).
        """
        requests = self._make_batch_delete_domains_requests(domains, batch_size=batch_size)
        async for message in self.call_stream(self.service, "batch_delete_domains", requests):  # pragma: no cover
            for domain in message or ():
                yield domain

    @staticmethod
    async def _make_manage_domain_state_flags_requests(
        domains: AnyIterable[Domain],
        state_flags: Dict[str, Optional[DatetimeRange]],
        close_requests: AnyIterable[str],
        *,
        batch_size: int = 100,
    ) -> AsyncIterable[ManageDomainStateFlagsRequest]:
        """Request generator for `manage_domain_state_flags` method."""
        async for chunk in chunked(domains, batch_size):
            request = ManageDomainStateFlagsRequest()
            for flag, validity in state_flags.items():
                validity_msg = request.state_flag_requests_to_create.state_flags.get_or_create(flag)
                if validity and validity.start:
                    validity_msg.valid_from.FromDatetime(validity.start)
                if validity and validity.end:
                    validity_msg.valid_to.FromDatetime(validity.end)
            for domain in chunk:
                message = ManageDomainStateFlagsRequest.Domain()
                message.domain_id.uuid.value = domain.domain_id
                if domain.domain_history_id:
                    message.domain_history_id.uuid.value = domain.domain_history_id
                request.state_flag_requests_to_create.domains.append(message)
            yield request

        async for close_chunk in chunked(close_requests, batch_size):  # pragma: no cover # because of py39 coverage
            request = ManageDomainStateFlagsRequest()
            for request_id in close_chunk:
                message = request.state_flag_requests_to_close.add()
                message.uuid.value = request_id
            yield request

    async def manage_domain_state_flags(
        self,
        *,
        domains: Optional[AnyIterable[Domain]] = None,
        state_flags: Optional[Mapping[str, Optional[DatetimeRange]]] = None,
        close_requests: Optional[AnyIterable[str]] = None,
        batch_size: int = 100,
    ) -> AsyncIterable[Dict[str, str]]:
        """Manage domains state flags.

        Args:
            domains: Iterable (or async iterable) of domains to be changed.
            state_flags: Dictionary. Keys are state flags, values are datetime ranges
                (or None if datetime range is from NOW to indefinitely).
            close_requests: String ids of domain state flag requests that should be closed.
            batch_size: Number of domains in single stream request.

        Returns:
            Async iterable of dictionaries with following keys:
                request_id: Contains ID of flag state request. It can be used to close the request later.
                domain_id: Domain ID.
                state_flag: Name of state flag that was set.

        Raises:
            DomainDoesNotExist: When domain does not exist.
            ValueError: When state flag or state flag request id does not exist.
        """
        if bool(domains) ^ bool(state_flags):
            raise ValueError("Either both `domains` and `state_flags` have to be defined or none of them.")

        requests = self._make_manage_domain_state_flags_requests(
            domains or [],
            cast(Dict[str, Optional[DatetimeRange]], state_flags) or {},
            close_requests or [],
            batch_size=batch_size,
        )
        async for message in self.call_stream(self.service, "manage_domain_state_flags", requests):  # pragma: no cover
            for record in message or []:
                record["domain_id"] = DomainId(record["domain_id"])
                yield record

    async def _get_domains_notify_info(
        self, request: Union[GetDomainsDeleteNotifyInfoRequest, GetDomainsOutzoneNotifyInfoRequest], method: str
    ) -> Dict[str, Any]:
        """Get domains notify info."""
        result: Dict[str, Any] = {"domains": []}
        async for message in self.call_stream(self.service, method, request):  # pragma: no cover
            if pagination := message.get("pagination"):
                result.setdefault("pagination", pagination)
            for domain in message.get("domains", []):
                domain["domain_id"] = DomainId(domain.get("domain_id"))
                domain["registrant_id"] = ContactId(domain.get("registrant_id"))
                result["domains"].append(domain)

        return result

    async def get_domains_outzone_notify_info(
        self,
        start: datetime,
        end: datetime,
        *,
        page_size: Optional[int] = None,
        page_token: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Get domains outzone notify info.

        Args:
            start: Start of the interval in which the domain is expected to stop being generated to zone.
            end: End of the interval in which the domain is expected to stop being generated to zone.
            page_token: Page token for pagination.
            page_size: Page size for pagination.
        """
        request = GetDomainsOutzoneNotifyInfoRequest()
        request.outzone_from.FromDatetime(start)
        request.outzone_to.FromDatetime(end)
        set_request_pagination(request, page_token=page_token, page_size=page_size)

        return await self._get_domains_notify_info(request, "get_domains_outzone_notify_info")

    async def get_domains_delete_notify_info(
        self,
        start: datetime,
        end: datetime,
        *,
        page_size: Optional[int] = None,
        page_token: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Get domains delete notify info.

        Args:
            start: Start of the interval in which the domain is expected to be deleted.
            end: End of the interval in which the domain is expected to be deleted.
            page_token: Page token for pagination.
            page_size: Page size for pagination.
        """
        request = GetDomainsDeleteNotifyInfoRequest()
        request.delete_from.FromDatetime(start)
        request.delete_to.FromDatetime(end)
        set_request_pagination(request, page_token=page_token, page_size=page_size)

        return await self._get_domains_notify_info(request, "get_domains_delete_notify_info")

    @staticmethod
    async def _make_update_domains_additional_notify_info_requests(
        request_type: Type[UpdateDomainsNotifyInfoRequest],
        info_message: Message,
        domains: Mapping[DomainId, Sequence[ContactInfo]],
        *,
        batch_size: int = 100,
    ) -> AsyncIterable[UpdateDomainsNotifyInfoRequest]:
        """Request generator for `update_domains_(outzone|delete)_additional_notify_info` methods.

        Args:
            request_type: Request message type.
            info_message: Additional info message with prefilled (outzone|delete)_from and (outzone|delete)_to fields.
            domains: Mapping of domains. Keys are domain ids, values are sequences of additional contact info.
            batch_size: Number of domains in single stream request.
        """
        async for chunk in chunked(domains.items(), batch_size):
            chunk_request = request_type()
            for domain_id, additional_info in chunk:
                domain = chunk_request.domains_additional_info.add()
                domain.CopyFrom(info_message)
                domain.domain_id.uuid.value = domain_id
                for contact in additional_info:
                    domain.additional_contacts.add(**{contact.type: contact.value})
            yield chunk_request

    async def update_domains_outzone_additional_notify_info(
        self,
        domains: Mapping[DomainId, Sequence[ContactInfo]],
        *,
        start: datetime,
        end: datetime,
        batch_size: int = 100,
    ) -> None:
        """Update additional notify info for domains that will stop being generated to zone.

        Args:
            domains: Mapping of domains. Keys are domain ids, values are sequences of additional contact info.
            start: Start of the interval in which the domain is expected to stop being generated to zone.
            end: End of the interval in which the domain is expected to stop being generated to zone.
            batch_size: Number of domains in single stream request.

        Raises:
            DomainDoesNotExist: When at least one of the domains does not exist.
            ValueError: When domain contact info format is invalid.
        """
        request = UpdateDomainsOutzoneAdditionalNotifyInfoRequest.DomainAdditionalInfo()
        request.outzone_from.FromDatetime(start)
        request.outzone_to.FromDatetime(end)

        await self.call(
            self.service,
            "update_domains_outzone_additional_notify_info",
            self._make_update_domains_additional_notify_info_requests(
                UpdateDomainsOutzoneAdditionalNotifyInfoRequest,
                request,
                domains,
                batch_size=batch_size,
            ),
        )

    async def update_domains_delete_additional_notify_info(
        self,
        domains: Mapping[DomainId, Sequence[ContactInfo]],
        *,
        start: datetime,
        end: datetime,
        batch_size: int = 100,
    ) -> None:
        """Update additional notify info for domains that are expected to be deleted.

        Args:
            domains: Mapping of domains. Keys are domain ids, values are sequences of additional contact info.
            start: Start of the interval in which the domain is expected to be deleted.
            end: End of the interval in which the domain is expected to be deleted.
            batch_size: Number of domains in single stream request.

        Raises:
            DomainDoesNotExist: When at least one of the domains does not exist.
            ValueError: When domain contact info format is invalid.
        """
        request = UpdateDomainsDeleteAdditionalNotifyInfoRequest.DomainAdditionalInfo()
        request.delete_from.FromDatetime(start)
        request.delete_to.FromDatetime(end)

        await self.call(
            self.service,
            "update_domains_delete_additional_notify_info",
            self._make_update_domains_additional_notify_info_requests(
                UpdateDomainsDeleteAdditionalNotifyInfoRequest,
                request,
                domains,
                batch_size=batch_size,
            ),
        )
